import sys
import argparse
from pathlib import Path
import copy
import pathlib

path_to_hermes = pathlib.Path(__file__).parent.resolve().joinpath("..")
sys.path.append(str(path_to_hermes))
from ..core import settings as st
from ..core.execution_directives import ExecDirectives

# todo: include the add to dictionary method from settings


def bool2str(s):
    if s.lower() in ["0", "false", "f"]:
        return False
    elif s.lower() in ["1", "true", "t"]:
        return True


class ExecutionFile:

    option_parser = argparse.ArgumentParser()
    option_parser.add_argument(
        "-ca",
        action="append",
        help="Add a file to be copied after execution",
        required=False,
    )
    option_parser.add_argument(
        "-cb",
        action="append",
        help="Add a file to be copied before execution",
        required=False,
    )
    option_parser.add_argument(
        "-ma",
        action="append",
        help="Add a file to be moved after execution",
        required=False,
    )
    option_parser.add_argument(
        "-l",
        "--log",
        type=bool2str,
        required=False,
        default="True",
        help="Store stdout and stderr in log files",
    )
    option_parser.add_argument(
        "-c",
        "--comment",
        action="append",
        help="Optional comment to be stored in hermes.info",
        required=False,
    )
    option_parser.add_argument(
        "-m",
        "--message",
        help="Send telegram message after job is done",
        default=False,
        action="store_true",
    )

    @staticmethod
    def write_pre(f=None, **pre_dict):
        """Write the PRE section of a hermes file given the info in pre_dict.
        If f is given write it to this file, else return a string.

        :param f: Optional file object
        :param pre_dict: A dictionary with entries "setting:value"
        """
        res = ["PRE"]
        for setting, value in pre_dict.items():
            res.append("%s:%s" % (setting, str(value)))
        res = "\n".join(res) + "\n"
        if f:
            f.write(res)
        else:
            return res

    @staticmethod
    def write_std(f=None, **std_dict):
        """Write the STD section of a hermes file given the info in std_dict.
        If f is given write it to this file, else return a string.

        :param f: Optional file object
        :param pre_dict: A dictionary with entries "flag:value"
        """
        res = ["STD"]
        for flag, value in std_dict.items():
            res.append("-%s %s" % (flag, str(value)))
        res = "\n".join(res) + "\n"
        if f:
            f.write(res)
        else:
            return res

    @staticmethod
    def write_exec(script_names, script_options_list, hermes_options_list, f=None):
        """Write the EXEC part of an .hermes file to f if given else return it a string.

        :param script_names: A list of script names as strings
        :param script_options_list: A list of dictionaries with entries "flag(str):value" describing script options.
               Flags should not include the '-' sign.
        :param hermes_options_list: A list of dictionaries with enrties "flag(str):value" describing hermes options.
               Flags should not include the '-' sign.
        :param f: (Optional) file object
        """
        res = ["EXEC"]
        line_format = "[[python, {scriptname}, {scriptoptions}],[{hermesoptions}]]"
        for scriptname, script_options, hermes_options in zip(
            script_names, script_options_list, hermes_options_list
        ):
            scriptoptions = ",".join(
                "-%s %s" % (flag, str(value)) for flag, value in script_options.items()
            )
            hermesoptions = ",".join(
                "-%s %s" % (flag, str(value)) for flag, value in hermes_options.items()
            )
            res.append(
                line_format.format(
                    scriptname=scriptname,
                    scriptoptions=scriptoptions,
                    hermesoptions=hermesoptions,
                )
            )
        res = "\n".join(res)
        if f:
            f.write(res)
        else:
            return res

    def get_std_args(self):
        return self.std_arguments

    def get_hermes_options(self, id):
        assert id < len(self.executions), (
            "Execution with id " + str(id) + "doesn't exist"
        )

        return self.executions[id][1]

    def get_exec_arguments(self, id):
        assert id < len(self.executions), (
            "Execution with id " + str(id) + "doesn't exist"
        )

        cmd = self.executions[id][0]

        if cmd[0].startswith("./"):
            args = cmd[1:]
        elif cmd[0].startswith("python"):
            args = cmd[2:]
        else:
            raise ValueError("Unexpected cmd in exec options: %s" % str(cmd))

        # add the std args
        args += self.std_arguments

        options = {}
        flags = []
        positional = []

        for arg in args:
            # positional argument
            if not arg.startswith("-"):
                positional.append(arg)
            else:
                split_arg = arg.split(" ")
                # flag
                if len(split_arg) == 1:
                    flags.append(arg)
                # option
                elif len(split_arg) == 2:
                    option, value = split_arg
                    options[option] = value
                # option with several args
                elif len(split_arg) > 2:
                    option = split_arg[0]
                    value = " ".join(split_arg[1:])
                    options[option] = value
                else:
                    raise ValueError("Could not parse exec argument %s" % arg)

        return positional, flags, options

    # make the magic happen -> check whether the option exists and parse it to the wanted type
    def parse_hermes_option(self, options):
        if not isinstance(options, list):
            options = [options]
        options = [opt for opt in options if opt != ""]
        if len(options) > 0:
            try:
                option_parser_input = [item.lstrip().rstrip() for item in options]
                option_parser_input = " ".join(option_parser_input)
                option_parser_input = option_parser_input.split(" ")
            except AttributeError:
                raise ValueError("Line: %s\n does not match EXEC convention")
            args = ExecutionFile.option_parser.parse_args(option_parser_input)
        else:
            args = ExecutionFile.option_parser.parse_args([])
        if not hasattr(args, "ca") or not args.ca:
            args.ca = []
        if not hasattr(args, "cb") or not args.cb:
            args.cb = []
        if not hasattr(args, "ma") or not args.ma:
            args.ma = []
        if not hasattr(args, "comment") or not args.comment:
            args.comment = []
        return args

    def parse_pre(self, lines):
        # parse the pre section of the file and update the settings
        for i, line in enumerate(lines):
            st.add_to_dictionary(self.settings, line, ":", i)

    def parse_std(self, stdh_lines, stde_lines):
        self.std_options = self.parse_hermes_option(stdh_lines)
        self.std_arguments = stde_lines

    def parse_exec_line(self, line):
        # split when closing the first bracket
        line_array = line.split("]")
        # remove the resulting empty strings and cast back to list
        line_array = list(filter(None, line_array))
        # make sure there are exactly two different elements left
        assert len(line_array) == 2, (
            "execution line did not fit our convention:\n%s" % line
        )

        # take cmd, delete opening braces
        cmd = line_array[0][2:]
        # parse to an actual list
        cmd = [itm.rstrip().lstrip() for itm in cmd.split(",")]

        # extract hermes options
        options_line = line_array[1].split(",")
        options_line = list(filter(None, options_line))

        options = [0 for _ in range(len(options_line))]
        for i, option in enumerate(options_line):
            tbrs = ["(", ")", "[", "]"]
            for tbr in tbrs:
                option = option.replace(tbr, "")
            options[i] = option

        options = self.parse_hermes_option(options)

        return cmd, options

    # so far, there is only the archive file option which can be added possibly multiple times
    # Thus, the union function only need to concatenate the options
    def union_options(self, standard_options, other_options):
        options = copy.deepcopy(standard_options)
        options.ca.extend(other_options.ca)
        options.cb.extend(other_options.cb)
        options.ma.extend(other_options.ma)
        options.comment.extend(other_options.comment)
        options.message = standard_options.message or other_options.message

        return options

    def parse_exec(self, lines):

        executions = [0 for _ in range(len(lines))]
        for i, line in enumerate(lines):
            cmd, options = self.parse_exec_line(line)

            options = self.union_options(self.std_options, options)

            executions[i] = (cmd, options)

        self.executions = executions

    def __init__(self, path_to_execution_file, comment=[], path_to_root=None):
        self.path_to_execution_file = (
            Path(path_to_execution_file)
            if isinstance(path_to_execution_file, str)
            else path_to_execution_file
        )
        self.settings = st.load(path_to_root=path_to_root)
        self.parse()
        for execution in self.executions:
            execution[1].comment.extend(comment)

    def __str__(self):
        return str(self.path_to_execution_file)

    def create_single_execution_hermesfile(self, name, num_execution):
        with open(name, "w") as f:
            # todo check whether the ending must be added
            for line in self.filelines[: self.exec_index + 1]:
                f.write(line + "\n")
            f.write(self.filelines[self.exec_index + num_execution + 1])

    def create_example_hermesfile(self, name="example.hermes"):
        pass
        # todo when format is really (!) finished )
        # else we will redo this a thousand times

    def parse(self):
        with open(self.path_to_execution_file, "r") as f:
            self.filelines = f.read().splitlines()

        # resolve exec directives into exec lines
        changes = True
        while changes:
            new_lines = []
            changes = False
            for line in self.filelines:
                compilation, had_effect = ExecDirectives.resolve(line)
                new_lines.extend(compilation)
                changes = changes or had_effect
            self.filelines = new_lines

        self.pre_index = self.filelines.index("PRE")
        self.stdh_index = self.filelines.index("STD-H")
        self.stde_index = self.filelines.index("STD-E")
        self.exec_index = self.filelines.index("EXEC")

        # parse pre
        self.parse_pre(self.filelines[self.pre_index + 1 : self.stdh_index])

        # parse std
        self.parse_std(
            self.filelines[self.stdh_index + 1 : self.stde_index],
            self.filelines[self.stde_index + 1 : self.exec_index],
        )

        # parse exec
        self.parse_exec(self.filelines[self.exec_index + 1 :])


if __name__ == "__main__":
    ef = ExecutionFile("example.hermes")
    ef.parse()



import os
import sys
import importlib
import shutil
import logging
from datetime import datetime
from pathlib import Path
import itertools
import contextlib
import time
import numpy as np
import torch
import pandas as pd
import matplotlib

# path_to_hermes = pathlib.Path(__file__).parent.resolve().joinpath("..")
# sys.path.append(str(path_to_hermes))
from .core import bot
from .core import settings as st
from .result import Result
from . import plot as hplt

import git


def _is_loaded(module_name: str) -> bool:
    return module_name in sys.modules


# https://stackoverflow.com/questions/14989858/get-the-current-git-hash-in-a-python-script
def get_current_commit_hash(default_commit="00000"):
    """Provided a repo return the hash of commit HEAD points to

    :param repo: git.Repo
    """
    try:
        repo = git.Repo(".", search_parent_directories=True)
    except:
        logging.warning("Unable to retrieve commit hash for repo head")
        return default_commit
    try:
        return repo.head.object.hexsha
    except:
        try:
            return repo.head.commit.hexsha
        except:
            logging.warning("Unable to retrieve commit hash for repo head")
            return default_commit


class Experiment:
    """Experiment class mainly taking care of storage. Every instance of the class will be unqiuely
    associated with a specific experiment directory and a script that was executed
    as part of the experiment.

    Experiment instances are used for the storage of important information, e.g. when it was executed,
    what were the command line arguments passed to the experiment script, what kind of data/results were generated
    during experiment execution, (...).

    It also provides ways of loading experiments from disk and methods for convenient access to the experiment
    information.
    """

    # FIELD NAME CONSTANTS
    COMMIT_HASH_N = "commit_hash"
    START_TIME_N = "start_time"
    END_TIME_N = "end_time"
    EXECUTION_TIME_N = "execution_time"
    COMMENT_N = "comments"

    # FILE NAME CONSTANTS
    INFO_FILE_NAME = "hermes.info"
    RE_EXECUTION_FN = "re_execution.hermes"
    SCRIPT_OPTIONS_FN = "exec_arguments"
    STDOUT_FN = "stdout.log"
    STDERR_FN = "stderr.log"

    TIME_FMT = "%Y_%m_%d_%H%M%S%f"

    def __init__(self, identifier, path_to_experiment, commit_hash, start_time):
        """This constructor should NEVER be used explicitly
        Use the classmethods Experiment.new when creating new experiments
        and Experiment.load when accessing finished experiments.

        :param identifier: the unqiue experiment identifier as a string
        :param path_to_experiment: the unique directory (as string or path-like object),
               built from the identifier, the script that was executed and the hermes experiments path.
        :param commit_hash: hash of the commit that was observed when running the script as string
        :param start_time: the time when this Experiment was created as datetime object
        """
        self.identifier = identifier
        self.path_to_experiment = Path(path_to_experiment)
        self.commit_hash = commit_hash
        self.start_time = start_time
        self.options = None

    @classmethod
    def new(cls, script_name, execution_file, job_idx, experiment_name):
        """
        **This method is used internally by CLI Hermes and not intended to be used outside of that scope.**

        Create a new Experiment instance for some job about to be executed

        :meta private:

        :param script_name: The name of the script the job takes care of as string
        :param execution_file: The ExecutionFile object (as created from some .hermes file) that is currently being handled
        :param job_idx: The execution index of the job as int
        :param experiment_name: name of the experiment as string. If provided this included in the experiment identifier
        """

        # get data relevant at start of experiment
        start_time = datetime.now()
        settings = execution_file.settings
        commit_hash = get_current_commit_hash()

        # check if experiments directory exists, if not make it
        path_to_experiments = Path(settings["path_to_experiments"])
        path_to_experiments = path_to_experiments.joinpath("experiments")
        if not os.path.exists(path_to_experiments):
            logging.info("Creating new directory %s" % path_to_experiments)
            os.mkdir(path_to_experiments)

        # check if [...]/experiments/script_name exists, if not create it
        script_name = script_name.replace(".py", "").lstrip().rstrip()
        path_to_script_directory = path_to_experiments.joinpath(script_name)
        if not path_to_script_directory.exists():
            logging.info("Creating script directory %s" % path_to_script_directory)
            try:
                os.mkdir(path_to_script_directory)
            except FileExistsError:
                # data race doesnt matter
                logging.warning(
                    "Script directory %s already exists, skipping creation."
                    % path_to_script_directory
                )

        # create directory [...]/experiments/script_name/experimentname
        path_to_experiment = path_to_script_directory.joinpath(experiment_name)
        if not path_to_experiment.exists():
            logging.info("Creating experiment directory %s" % path_to_experiment)
            os.mkdir(path_to_experiment)
            try:
                os.mkdir(path_to_experiment)
            except FileExistsError:
                # data race doesnt matter
                logging.warning(
                    "Experiment directory %s already exists, skipping creation"
                    % path_to_experiment
                )

        # create directory [...]/experiments/script_name/experimentname/date_commitnumber[4]
        # if it already exists throw an error
        experiment_identifier = [
            start_time.strftime(Experiment.TIME_FMT),
            commit_hash[-5:],
        ]
        experiment_identifier = "_".join(experiment_identifier)
        path_to_experiment = path_to_experiment.joinpath(experiment_identifier)
        if path_to_experiment.exists():
            logging.error(
                "Cannot create new Experiment - Experiment Directory already exists: %s",
                path_to_experiment,
            )
        logging.info("Creating job directory %s", path_to_experiment)
        os.mkdir(path_to_experiment)

        experiment = cls(
            experiment_identifier, path_to_experiment, commit_hash, start_time
        )
        options = execution_file.get_hermes_options(job_idx)
        experiment.set_options(options)

        # store execution file in info
        reexec_file_destination = experiment.file(Experiment.RE_EXECUTION_FN)
        execution_file.create_single_execution_hermesfile(
            reexec_file_destination, job_idx
        )

        # store commit hash in info
        experiment._add_to_info(Experiment.COMMIT_HASH_N, commit_hash)

        # copy before execution
        experiment._copy_files_before_execution()

        # store options given in script
        args = execution_file.get_exec_arguments(job_idx)
        experiment.store_script_arguments(*args)

        # touch log files
        experiment.file(Experiment.STDOUT_FN).touch()
        experiment.file(Experiment.STDERR_FN).touch()

        return experiment

    @classmethod
    def load(cls, path_to_experiment):
        """
        Create an Experiment object for a finished experiment

        :param path_to_experiment:
        :type path_to_experiment: Union[str, os.PathLike]
        """
        assert Experiment.is_experiment(
            path_to_experiment
        ), "The given path does not store any hermes experiment"
        p = Path(path_to_experiment)
        script_name, experiment_name, identifier = p.parts[-3:]
        info = Experiment.read_info(path_to_experiment)
        commit_hash = info["commit_hash"]
        try:
            start_time = time.strptime(info["start_time"], Experiment.TIME_FMT)
        except:
            try:
                start_time = float(info["start_time"])
                logging.info(
                    "Failed to load experiment start time as time.struct_time, loaded as floating point number instead."
                )
            except:
                start_time = None
                logging.warning("Failed to load experiment start time.")

        experiment = cls(identifier, path_to_experiment, commit_hash, start_time)

        return experiment

    @property
    def exec_args(self):
        """
        A set of the execution arguments that were stored for this experiment, e.g.
        {"-lr 0.001", "--gradclip"}

        :return: The set of execution arguments
        :rtype: set[str]
        """
        if hasattr(self, "_exec_args"):
            return self._exec_args
        else:
            with open(
                self.path_to_experiment.joinpath(Experiment.SCRIPT_OPTIONS_FN), "r"
            ) as f:
                lines = f.readlines()
            lines = [
                line.rstrip().lstrip()
                for line in lines
                if len(line.replace(" ", "")) > 0
            ]
            self._exec_args = set(lines)
            return self._exec_args

    @property
    def exec_args_dict(self):
        """
        A dictionary mapping execution argument names to values as specified in the execution file used to run this experiment.

        :return: The execution argument dictionay
        :rtype: dict[str, Any]
        """
        if hasattr(self, "_exec_args_dict"):
            return self._exec_args_dict
        else:
            self._exec_args_dict = {}
            for arg in self.exec_args:
                aux = arg.split(" ")
                if len(aux) <= 1:
                    key = aux[0]
                    val = True
                elif len(aux) == 2:
                    key, val = aux
                    try:
                        val = float(val)
                    except:
                        val = str(val)
                elif len(aux) > 2:
                    logging.warning("Unknown exec argument format: %s" % arg)
                    key = aux[0]
                    vals = aux[1:]
                    try:
                        val = [int(val) for val in vals]
                    except:
                        try:
                            val = [float(val) for val in vals]
                        except:
                            val = [str(val) for val in vals]
                if key in self._exec_args_dict:
                    try:
                        self._exec_args_dict[key].append(val)
                    except:
                        self._exec_args_dict[key] = [self._exec_args_dict[key], val]
                else:
                    self._exec_args_dict[key] = val
            return self._exec_args_dict

    @property
    def scores_file(self):
        """
        Returns the path to the scores file stored in this experiment as a pathlib.Path
        """
        return self.file("%s.scores" % self.identifier)

    def file(self, file):
        """

        :param file: Name of a file
        :type file: Union[str, os.PathLike]

        :return: the path to the file in the experiment directory
        :rtype: pathlib.Path
        """
        return self.path_to_experiment.joinpath(file)

    def load_state_dict(self, pth_name, net=None):
        """
        Load a state dictionary

        :param pth_name: .pth state dict file name
        :type pth_name: str

        :param net: optionally load state dict into this network module
        :type net: torch.nn.Module, optional
        """
        pth_file = self.file(pth_name)

        try:
            device = net.device
        except:
            device = torch.device("cpu")

        try:
            state_dict = torch.load(pth_file, map_location=device)
        except FileNotFoundError:
            logging.critical("Unable to find this pth file.")
            return None
        if net is None:
            return state_dict
        else:
            net.load_state_dict(state_dict)
            net.eval()
            return net

    def list_pth_files(self):
        """
        Return a list of all .pth files (i.e. pytorch state dictionaries) within the experiment directory

        :return: the list
        :rtype: list[str]
        """
        return list(pth_path.name for pth_path in self.path_to_experiment.glob("*.pth"))

    def scores(self):
        """
        Loads the scores from the experiments' scores file

        :return: The scores
        :rtype: numpy.ndarray
        """
        try:
            with open(self.scores_file, "r") as f:
                return np.array([float(line) for line in f])
        except FileNotFoundError:
            logging.error(
                "Tried to load scores, but not scores file %s was not found."
                % self.scores_file
            )

    @property
    def result(self):
        """
        Load the pyhermes result associated with this experiment, if it exists
        :return: pyhermes.result.Result
        :raises RuntimeError: if the experiment does not contain a pyhermes result
        """
        if hasattr(self, "_result"):
            return self._result
        else:
            result_files = list(self.path_to_experiment.glob("*.result"))
            if len(result_files) == 0:
                raise RuntimeError("Experiment does not contain a pyhermes result")
            if len(result_files) > 1:
                logging.warning(
                    "Experiment contains more than one result file, loading the first one only"
                )
            self._result = Result.load(result_files[0])
            return self._result

    def load_network_class(self, keys=["--neural_network_file", "-nnf"], default="fcn"):
        # always extend path to experiment directory to load copied files from there.
        path = str(self.path_to_experiment.absolute())
        sys.path.append(path)
        if all(key not in self.exec_args_dict for key in keys):
            logging.warning(
                "Experiment at path %s does not specify neural network file in exec args (given keys %s). Using default network name."
                % (self.path_to_experiment, str(keys))
            )
            name = default
            network_module = importlib.import_module(name)
            return network_module.Network

        network_classes = dict()
        for key in keys:
            if key not in self.exec_args_dict:
                continue
            nnf_string = self.exec_args_dict[key]
            if not nnf_string.endswith(".py"):
                nnf_string += ".py"
            nnf = Path(nnf_string)
            possible_paths = [nnf]
            try:
                nnf_string_short = nnf_string.split(".")[-2] + ".py"
                possible_paths.append(Path(nnf_string_short))
            except:
                pass

            for possible_path in possible_paths:
                print(possible_path)

                try:
                    network_module = importlib.import_module(possible_path.stem)
                    network_classes[key] = network_module.Network
                except:
                    continue

        if len(network_classes) == 1:
            return list(network_classes.values())[0]
        elif len(network_classes) == 0:
            logging.warning(
                "Unable to load any network classes from %s." % self.path_to_experiment
            )
            return None
        else:
            return network_classes

    @staticmethod
    def is_experiment(path):
        """
        Check if a given directory represents a hermes experiment.

        :param path: path to the directory
        :type path: Union[str, os.PathLike]

        :return: whether the path represents an experiment
        :rtype: bool
        """
        p = Path(path)
        assert p.exists(), "Path must exist"
        return p.is_dir() and len(list(p.glob("hermes.info"))) >= 1

    @staticmethod
    def load_all(path):
        """
        Iterate over all subdirectories of a given parent directory and load all experiments from present experiment directories

        :param path: The parent directory.

        :return: A list of experiment
        :rtype: ExperimentList
        """
        p = Path(path)
        return ExperimentList(
            Experiment.load(subpath)
            for subpath in p.iterdir()
            if Experiment.is_experiment(subpath)
        )

    @staticmethod
    def read_info(path_to_experiment):
        """
        Try to read the `hermes.info` of a hermes experiment directory and return it as dictionary

        :param path_to_experiment: path to the experiment directory
        :type path_to_experiment: Union[str, os.PathLike]

        :return: experiment info
        :rtype: dict[str, str]
        """
        with open(os.path.join(path_to_experiment, "hermes.info"), "r") as f:
            info = dict()
            for line in f.readlines():
                key, val = line.split(":")
                info[key] = val
        return info

    @property
    def info_dict(self):
        """
        Similar to :func:`~Experiment.read_info`, but as instance property rather than staticmethod

        :return: experiment info
        :rtype: dict[str, str]
        """
        if hasattr(self, "_info_dict"):
            return self._info_dict
        else:
            self._info_dict = Experiment.read_info(self.path_to_experiment)
            return self._info_dict

    @contextlib.contextmanager
    def open_log_files(self, mode="r+"):
        """
        Context manager for opening the STDOUT and STDERR log files.

        Usage:

        .. code-block:: python

           from pyhermes.storage import Experiment
           exp = Experiment.load(...)
           with exp.open_log_files(mode='r') as f_stdout, f_stderr:
              # do stuff

        :param mode: The filemode e.g. 'w', 'a', ... Defaults to 'r+'.
        """
        f_stdout = open(self.file(Experiment.STDOUT_FN), mode)
        f_stderr = open(self.file(Experiment.STDERR_FN), mode)
        yield f_stdout, f_stderr
        f_stdout.close()
        f_stderr.close()

    @contextlib.contextmanager
    def open_stderr_file(self, mode="r"):
        """
        Context manager for opening the STDERR file only

        :param mode: The filemode e.g. 'w', 'a', ... Defaults to 'r+'.
        """
        f_stderr = open(self.file(Experiment.STDERR_FN), mode)
        yield f_stderr
        f_stderr.close()

    def store_script_arguments(self, positionals, flags, options, mode="w"):
        """
        Store the scripts options/parameters (e.g. -lr 0.005) in human-readable form

        :param positionals:
        :param flags:
        :param options: a dictionary created by ExecutionFile.get_job_options

        :meta private:
        """
        with open(self.file(Experiment.SCRIPT_OPTIONS_FN), mode) as f:
            for positional in positionals:
                f.write("%s\n" % positional)
            for flag in flags:
                f.write("%s\n" % flag)
            for option, value in options.items():
                f.write("%s %s\n" % (option, value))

    def append_script_argument(self, positionals=None, flags=None, options=None):
        self.store_script_arguments(positionals, flags, options, mode="a")

    def _copy_files_before_execution(self):
        """
        Called by Experiment.new. Copies files specified with -cb hermes option
        :meta private:
        """
        for file_name in self.options.cb:
            self.copy_file(file_name)

    def _copy_files_after_execution(self):
        """
        Called by Experiment.report. Copies files specified with -ca hermes option
        :meta private:
        """
        for file_name in self.options.ca:
            self.copy_file(file_name)

    def _move_files_after_execution(self):
        """
        Called by Experiment.report. Moves files specified with -ma hermes option
        :meta private:
        """
        for file_name in self.options.ma:
            self.move_file(file_name)

    def store_comments(self):
        """
        Store comments given by -c option either through main call or hermes option
        :meta private:
        """
        self._add_to_info(
            Experiment.COMMENT_N, ",".join(cmt for cmt in self.options.comment)
        )

    def set_options(self, options):
        """Setter for hermes options. Called by Experiment.new

        :param options: A Namespace object as returned by ExecutionFile.hermes_option_parser and Experiment.union_options
        :meta private:
        """
        self.options = options

    def get_identifier(self):
        """

        :return: the unique hermes identifier assigned to this experiment
        :rtype: str
        """
        return self.identifier

    def copy_file(self, path_to_file, new_file_name=None):
        """Method to copy a file to the unique experiment directory.

        :param path_to_file: A string or path-like object
        :param new_file_name: Optional new name for the copy. If None, the basename of
               path_to_file is used.
        :meta private:
        """
        file_name = new_file_name if new_file_name else os.path.basename(path_to_file)
        if not os.path.exists(path_to_file):
            logging.error(
                "File to be copied to experiment directory %s does not exist: %s",
                self.path_to_experiment,
                path_to_file,
            )
        shutil.copyfile(path_to_file, self.file(file_name))

    def move_file(self, path_to_file, new_file_name=None):
        """Method to move a file to the unique epxeriment directory.

        :param path_to_file: A string or path-like object
        :param new_file_name: Optional new name for the moved file. If None, the basename of
               path_to_file is used.
        :meta private:
        """
        file_name = new_file_name if new_file_name else os.path.basename(path_to_file)
        if not os.path.exists(path_to_file):
            logging.error(
                "File to be moved to experiment directory %s does not exist: %s",
                self.path_to_experiment,
                path_to_file,
            )
        shutil.move(path_to_file, self.file(file_name))

    def _add_to_info(self, key, value):
        """Internal method to add a line "key:value" to the hermes.info file.
        This file captures general information such as execution time, start time, ...

        :param key: The name of the object to be stored as string
        :param value: The object to be stored. Any object that implements __str__
        :meta private:
        """
        with open(self.file(Experiment.INFO_FILE_NAME), "a") as f:
            f.write("%s:%s\n" % (key, str(value)))

    def store_times(self, start_time, end_time):
        """Internal method that stores job start time, end time and execution time (i.e. difference of the former two)

        :param start_time: System time [s] as measured just before job is launched
        :param end_time: System time [s] as measured just after job terminates
        :meta private:
        """
        execution_time = end_time - start_time
        start_time = time.localtime(start_time)
        end_time = time.localtime(end_time)
        self._add_to_info(
            Experiment.START_TIME_N, time.strftime(Experiment.TIME_FMT, start_time)
        )
        self._add_to_info(
            Experiment.END_TIME_N, time.strftime(Experiment.TIME_FMT, end_time)
        )
        self._add_to_info(Experiment.EXECUTION_TIME_N, str(execution_time))

    def _fetch_new_files(self):
        """
        Scan the working directory for files containing the experiment identifier
        and move them to the experiment directory
        :meta private:
        """
        dir_list = os.listdir(".")
        for part in dir_list:
            # skip parts of the directory that don't contain the idx
            if self.identifier not in part:
                continue
            self.move_file(part)

    def telegram_notifications(self, start_time, end_time):
        if self.options.message:
            self.bot_handler = bot.Bot_handler()

            message = "Job " + str(self.identifier) + " was completed\n"

            if len(self.exec_args) != 0:
                message += "Options: "
                for opt in self.exec_args:
                    message += str(opt) + " "
                message += "\n"
            else:
                message += "No further options were specified\n"

            message += "Duration: %.2f seconds" % (end_time - start_time)

            if len(self.scores()) != 0:
                pic_path = "tmp_trainingplot.png"
                matplotlib.use("agg")
                hplt.plot_training_progress_from_experiment(self, path=pic_path)

                self.bot_handler.picture(pic_path, caption=message)
                os.remove(pic_path)
            else:
                self.bot_handler.message(message)

    def _report(self, start_time, end_time):
        """
        Called by Execution once the job for this experiment is finished.
        Used to transmit job execution info and makes sure all storage operations
        scheduled to take place after execution are carried out.

        :param start_time: Job start_time
        :param end_time: Job end_time
        :meta private:
        """
        self.store_times(start_time, end_time)
        self._copy_files_after_execution()
        self._move_files_after_execution()
        self._fetch_new_files()
        self.store_comments()
        self.telegram_notifications(start_time, end_time)

    def erase(self):
        logging.info("Erasing job directory %s", self.path_to_experiment)
        shutil.rmtree(self.path_to_experiment)


class ExperimentList(list):
    """
    A wrapper around :python:`list`. Offers some useful functions for selecting/filtering
    loaded experiments. Will be extended in the future.

    These lists are typically created by :meth:`pyhermes.storage.Experiment.load_all`, but you can
    also create them in your own.
    """

    def __init__(self, exp_iterable):
        """
        Create an ExperimentList from given experiments

        :param exp_iterable: an iterable collection of experiments
        :type exp_iterable: Iterable[pyhermes.storage.Experiment]
        """
        super(ExperimentList, self).__init__(exp_iterable)
        assert all(isinstance(itm, Experiment) for itm in self)

    def filter_by_exec_args(self, *args):
        """
        :param args: Argument as strings
        :return: A new experiment list containing only those experiments which have the given args in their exec args set.

        Usage example:

        .. code-block:: python

           from pyhermes.Storage import Experiment
           path = "some path where many experiments are stored"
           experiments = Experiment.load_all(path)
           ddqn_runs = Experiment.filter_by_exec_args("--doubledqn")

        """
        args = set(args)
        return ExperimentList(exp for exp in self if args < exp.exec_args)

    def latest(self):
        """
        This method makes use of the fact hermes tracks the start time of each experiment,

        :return: The most recent experiment within the list
        :rtype: Experiment
        """
        return min(self, key=lambda exp: exp.start_time)


if _is_loaded("pandas") or _is_loaded("pd"):

    class _ExperimentFrame(pd.DataFrame):
        """
        Test doc
        """

        def __init__(self, exp_iterable):
            experiments = list(exp_iterable)
            assert all(isinstance(exp, Experiment) for exp in experiments)
            super(ExperimentFrame, self).__init__(index=range(len(experiments)))

            self["experiments"] = experiments
            union_exec_arg_keys = set(
                itertools.chain(
                    *(list(exp.exec_args_dict.keys()) for exp in experiments)
                )
            )
            for key in union_exec_arg_keys:
                self[key] = [
                    exp.exec_args_dict[key] if key in exp.exec_args_dict else None
                    for exp in experiments
                ]


else:

    class _ExperimentFrame:
        def __init__(self, *args, **kwargs):
            raise RuntimeError("ExperimentFrame class needs pandas as requirement!")


ExperimentFrame = _ExperimentFrame



import sys
import time
import subprocess
import os
import threading
import logging
import pathlib

path_to_hermes = pathlib.Path(__file__).parent.resolve().joinpath("..")
sys.path.append(str(path_to_hermes))
from ..storage import Experiment


class Execution:
    def worker(self, id):
        while True:
            # get the job number
            self.lock.acquire()
            current_job = self.finished_jobs
            self.finished_jobs += 1
            self.lock.release()

            # check whether all jobs are done
            if current_job >= self.job_number:
                return

            # get the job and extract the command
            cmd, options = self.jobs[current_job]

            # extract script_name. If started with ./  its the first word of the command
            # else it is the second one
            if cmd[0][:2] == "./":
                script_name = cmd[0][2:]
                option_start = 1
            else:
                script_name = os.path.basename(cmd[1])
                option_start = 2

            ex = Experiment.new(
                script_name, self.execution_file, current_job, self.experiment_name
            )
            identifier = ex.get_identifier()

            # rebuild command: 1. script start 2. hermes_name 3. std arguments 4. current arguments
            if not os.path.exists(cmd[option_start - 1]):
                raise FileNotFoundError("Script %s not found" % cmd[option_start - 1])

            cmd = (
                cmd[:option_start]
                + [identifier]
                + self.execution_file.get_std_args()
                + cmd[option_start:]
            )
            cmd = " ".join(cmd)

            start = time.time()
            logging.info('Launching job Nr. %d: "%s".', current_job, cmd)
            logging.info(
                "STDOUT and STDERR will be written to %s.",
                "log files" if options.log else "console",
            )
            if options.log:
                with ex.open_log_files() as (f_stdout, f_stderr):
                    completed_process = subprocess.run(
                        cmd, stdout=f_stdout, stderr=f_stderr, shell=True
                    )
            else:
                completed_process = subprocess.run(cmd, shell=True)
            end = time.time()

            if completed_process.returncode == 0:
                logging.info(
                    "Sucessfully finished job Nr. %d %s in %.2f seconds"
                    % (current_job, identifier, (end - start))
                )
            else:
                logging.critical(
                    "Job %s (Nr. %d) terminated after %.2f seconds with exit status %d"
                    % (
                        identifier,
                        current_job,
                        (end - start),
                        completed_process.returncode,
                    )
                )
                if options.log:
                    logging.info("Here is the jobs stderr:")
                    with ex.open_stderr_file() as f_stderr:
                        print(f_stderr.read())

            ex._report(start, end)

            # when debugging we might want to not save the experiment results
            # to prevent clutter and save disk space
            if self.debug:
                ex.erase()

    def run(self):
        start = time.time()
        threads = [
            threading.Thread(target=self.worker, args=[i])
            for i in range(self.num_threads)
        ]
        self.lock = threading.Lock()

        for t in threads:
            t.start()

        for t in threads:
            t.join()

        end = time.time()

    def __init__(self, execution_file, experiment_name=None, debug=False):
        self.execution_file = execution_file
        self.jobs = execution_file.executions
        self.job_number = len(self.jobs)
        self.settings = execution_file.settings
        self.num_threads = self.settings["threads"]
        self.finished_jobs = 0
        self.experiment_name = experiment_name
        self.debug = debug

    def print(self):
        for job in self.jobs:
            print(job)



import numpy as np
from collections import deque
import matplotlib.pyplot as plt


def smoothen(arr, window_size=100):
    """Function that computes the sliding mean over the given array.
        For the first numbers n < window_size, the sliding mean is computed using the numbers there are.

    Args:
        arr (float array): scores
        window_size (int, optional): size of sliding window. Defaults to 100.

    Returns:
        (float array, float arrary): list with sliding means, list with stds
    """
    means = []
    stds = []
    for i in range(len(arr)):
        lower = np.maximum(0, i - window_size)
        means.append(arr[lower : i + 1].mean())
        stds.append(arr[lower : i + 1].std())
    return means, stds


def plot_training_progress(
    scores,
    eps_start=None,
    eps_end=None,
    eps_decay=None,
    epsilons=None,
    window_length=250,
    figsize=(12, 12),
    fontsize=18,
    path=None,
    scores_color="c",
    mean_color="r",
    score_range=None,
    epsilon_range=None,
    axis_gap=0.02,
    mark_best=False,
    legend=True,
    ax=None,
    suppress_single_scores=False,
):
    """Plot the training curve given an array of training scores, i.e. training episode returns

    Args:
        scores (float array): scores, i.e. training episode returns
        eps_start (int, optional): epsilon-greedy exploration coefficient in the training beginning. Defaults to 1.
        eps_end (float, optional): epsilon-greedy exploration coefficient in the end of training. Defaults to 0.001.
        eps_decay (float, optional): epsilon-greedy exponential decay factor from eps_start to eps_end. Defaults to 0.999.
        epsilons (list, optional): alternative to eps_start, eps_end and eps_decay. Provide all used eps values. Defaults to []. Will be ignored, if eps_start, eps_end and eps_decay are given
        window_length (int, optional): size of window_length for sliding window. Defaults to 250.
        figsize (tuple, optional): Figure size to use.. Defaults to (12,12).
        fontsize (int, optional): fontsize to use. Defaults to 18.
        path (string, optional): path to save the created plot. Defaults to None.
        scores_color (str, optional): score dot color. Defaults to 'c'.
        mean_color (str, optional): slidign mean color. Defaults to 'r'.
        score_range ((float, float), optional): range to be shown in the plot on the y-axis. Defaults to None.
        epsilon_range ((float, float)), optional): range to be shown in the plot in the epsilon y-axis. Defaults to None.
        axis_gap (float, optional):  gap to be added to score_range and epsilon_range for better readability. Defaults to 0.02.
        mark_best (bool, optional): whether the best observed value should be marked. Defaults to False.
        legend (bool, optional): whether to plot a legend. Defaults to True.
        ax (pyplot axis, optional): pyplot axis to use. Generating a new one if none provided. Defaults to None.
        suppress_single_scores (bool, optional): whether to suppress the single score dots. Defaults to False.

    Returns:
        plt: generated plot.
    """
    # check whether the epsilons are explicitly given and calculate them otherwise
    if (eps_start is not None) or (eps_end is not None) or (eps_decay is not None):
        assert all(
            arg is not None for arg in [eps_start, eps_end, eps_decay]
        ), "You need to provide all of eps_start, eps_end and eps_decay if you want to plot the effect of the epsilon-greedy policy"
        epsilons = []
        for i in range(len(scores)):
            actual_eps = np.max([eps_start * np.power(eps_decay, i), eps_end])
            epsilons.append(actual_eps)

    if epsilons is not None and len(epsilons) > 0:
        assert len(epsilons) == len(
            scores
        ), "The length of the list of epsilon (exploration coefficient) values need to be the same as the amount of scores"

    plot_epsilon = epsilons is not None and len(epsilons) > 0

    # calculate the mean by using the given window_length
    scores_window = deque(maxlen=window_length)

    means = []
    for i, score in enumerate(scores):
        scores_window.append(score)
        means.append(np.mean(scores_window))

    if ax == None:
        # create the figure with the given figsize
        fig, ax1 = plt.subplots(figsize=figsize)
    else:
        ax1 = ax

    ax1.set_ylabel("Return", fontsize=fontsize)
    ax1.set_xlabel("Episode #", fontsize=fontsize)

    x = range(len(scores))

    # plt.xticks(fontsize = fontsize)

    # scatter the scores and the means on one axis
    if not suppress_single_scores:
        ax1.scatter(x, scores, label="Episodes", color=scores_color, alpha=0.8, s=1)
    ax1.plot(range(len(means)), means, label="Sliding Mean", color=mean_color)

    if plot_epsilon:
        # optionally scatter epsilon on the other axis
        ax2 = ax1.twinx()
        ax2.set_ylabel("Random Exploration Coefficient", fontsize=fontsize)
        ax2.plot(x, epsilons, label="Epsilon")
        ax2.axis([0, len(scores), -0, 1])

    # set the given score range or calculate it relative to the given values
    if score_range == None:
        score_min = np.min(scores)
        score_max = np.max(scores)
        score_range = np.abs(score_max - score_min)
        score_min = score_min - axis_gap * score_range
        score_max = score_max + axis_gap * score_range
    else:
        score_min, score_max = score_range
    ax1.set_ylim(score_min, score_max)

    # do the same for epsilon
    if plot_epsilon:
        if epsilon_range == None:
            epsilon_min = eps_end
            epsilon_max = eps_start
            epsilon_range = np.abs(epsilon_max - epsilon_min)
            epsilon_min = epsilon_min - axis_gap * epsilon_range
            epsilon_max = epsilon_max + axis_gap * epsilon_range
        else:
            epsilon_min, epsilon_max = epsilon_range

        ax2.set_ylim(epsilon_min, epsilon_max)
        ax2.tick_params(labelsize=fontsize)

    # set the fontsize of the axis/plt
    ax1.axis()
    plt.axis()
    ax1.tick_params(labelsize=fontsize)

    # ax1.legend(bbox_to_anchor=(0.8, 0.2, 0.15, .102), fontsize = 16, mode = 'expand', ncol=1, handlelength = 7, numpoints = 10)
    # ax1.legend(fontsize = fontsize, ncol = 2, bbox_to_anchor=(0.05, 0.8, 0.9, 0.12), mode = 'expand', scatterpoints=5)
    # ax2.legend(fontsize = fontsize, bbox_to_anchor=(0.675, 0.9, 0.3, 0.12), mode = 'expand')
    if mark_best:
        best = np.argmax(means)
        best_value = -float("inf")
        other_best = None
        plt.axvline(
            x=best, linestyle="dotted", color="#A0A0A0", label="Highest overall result"
        )
        start_id = 0
        while start_id < len(scores):
            value = np.mean(scores[start_id : start_id + window_length])
            if value > best_value:
                other_best = start_id + window_length - 1
                best_value = value
            start_id += window_length

        plt.axvline(x=other_best, linestyle="dotted", color="k", label="Best policy")

    if legend:
        ax1.legend(fontsize=fontsize)
        # ax2.legend()

    if path != None:
        assert ax == None
        plt.savefig(path, bbox_inches="tight")

    return plt


def plot_training_progress_from_file(
    file,
    eps_start=1,
    eps_end=0.001,
    eps_decay=0.999,
    epsilons=[],
    window_length=250,
    figsize=(12, 12),
    fontsize=18,
    path=None,
    scores_color="c",
    mean_color="r",
    score_range=None,
    epsilon_range=None,
    axis_gap=0.02,
    mark_best=True,
    legend=True,
    suppress_singe_scores=False,
):
    """Read the trainings scores from file and plot the training progress.

    Args:
        file (string): path to file
        eps_start (int, optional): exploration coefficient in the training beginning. Defaults to 1.
        eps_end (float, optional): exploration coefficient in the end of training. Defaults to 0.001.
        eps_decay (float, optional): exponential decay factor from eps_start to eps_end. Defaults to 0.999.
        epsilons (list, optional): alternative to eps_start, eps_end and eps_decay. Provide all used eps values. Defaults to [].
        window_length (int, optional): size of window_length for sliding window. Defaults to 250.
        figsize (tuple, optional): Figure size to use.. Defaults to (12,12).
        fontsize (int, optional): fontsize to use. Defaults to 18.
        path (string, optional): path to save the created plot. Defaults to None.
        scores_color (str, optional): score dot color. Defaults to 'c'.
        mean_color (str, optional): slidign mean color. Defaults to 'r'.
        score_range ((float, float), optional): range to be shown in the plot on the y-axis. Defaults to None.
        epsilon_range ((float, float)), optional): range to be shown in the plot in the epsilon y-axis. Defaults to None.
        axis_gap (float, optional):  gap to be added to score_range and epsilon_range for better readability. Defaults to 0.02.
        mark_best (bool, optional): whether the best observed value should be marked. Defaults to False.
        legend (bool, optional): whether to plot a legend. Defaults to True.
        suppress_single_scores (bool, optional): whether to suppress the single score dots. Defaults to False.


    Returns:
        pyplot: Plot
    """
    scores = []
    with open(file, "r") as f:
        first = f.readline()
        try:
            first = float(first)
            scores.append(first)
        except:
            print("Dismiss first line")

        for line in f:
            score = float(line)
            scores.append(score)

    return plot_training_progress(
        scores,
        eps_start,
        eps_end,
        eps_decay,
        epsilons,
        window_length,
        figsize,
        fontsize,
        path,
        scores_color,
        mean_color,
        score_range,
        epsilon_range,
        axis_gap,
        mark_best=mark_best,
        legend=legend,
    )


# todo: this doesn't feel right. the information is encoded in the experiment and should be read from there.
def plot_training_progress_from_experiment(
    exp,
    eps_start_key="-es",
    eps_end_key="-ee",
    eps_decay_key="-ed",
    window_length=250,
    figsize=(12, 12),
    fontsize=18,
    path=None,
    scores_color="c",
    mean_color="r",
    score_range=None,
    epsilon_range=None,
    axis_gap=0.02,
    mark_best=True,
    legend=True,
    suppress_single_scores=False,
):
    """Read the trainings scores from file and plot the training progress.

    Args:
        exp (Experiment): experiment
        window_length (int, optional): size of window_length for sliding window. Defaults to 250.
        figsize (tuple, optional): Figure size to use.. Defaults to (12,12).
        fontsize (int, optional): fontsize to use. Defaults to 18.
        path (string, optional): path to save the created plot. Defaults to None.
        scores_color (str, optional): score dot color. Defaults to 'c'.
        mean_color (str, optional): slidign mean color. Defaults to 'r'.
        score_range ((float, float), optional): range to be shown in the plot on the y-axis. Defaults to None.
        epsilon_range ((float, float)), optional): range to be shown in the plot in the epsilon y-axis. Defaults to None.
        axis_gap (float, optional):  gap to be added to score_range and epsilon_range for better readability. Defaults to 0.02.
        mark_best (bool, optional): whether the best observed value should be marked. Defaults to False.
        legend (bool, optional): whether to plot a legend. Defaults to True.
        suppress_single_scores (bool, optional): whether to suppress the single score dots. Defaults to False.

    Returns:
        pyplot: Plot
    """
    file = exp.scores_file

    def _get(*keys):
        return [exp.exec_args_dict[key] for key in keys]

    try:
        eps_start, eps_end, eps_decay = _get(eps_start_key, eps_end_key, eps_decay_key)
    except KeyError:
        print("Not plotting epsilon", exp.exec_args_dict)
        eps_start = None
        eps_end = None
        eps_decay = None

    epsilons = []

    return plot_training_progress_from_file(
        file,
        eps_start,
        eps_end,
        eps_decay,
        epsilons,
        window_length,
        figsize,
        fontsize,
        path,
        scores_color,
        mean_color,
        score_range,
        epsilon_range,
        axis_gap,
        mark_best=mark_best,
        legend=legend,
    )



import sys
import argparse


class Argument_parser:
    def __init__(self):
        self.parser = argparse.ArgumentParser(
            description="Process the common RL arguments"
        )

        # General Arguments
        self.parser.add_argument(
            "hermes_name",
            type=str,
            help="the unique name determined by the hermes programm",
        )
        self.parser.add_argument("-s", "--seed", help="seed", default=0, type=int)

        # DQN Arguments

        # # DQN binaries
        self.parser.add_argument(
            "-gpu",
            "--gpu",
            help="using gpu instead of CPU, if possible",
            default=False,
            action="store_true",
        )
        self.parser.add_argument(
            "-erb",
            "--extract_replay_buffer",
            help="extract all states added to the rb",
            default=False,
            action="store_true",
        )

        # # DQN options
        self.parser.add_argument(
            "-ne",
            "--num_episodes",
            help="number of episodes to learn from",
            default=10000,
            type=int,
        )
        self.parser.add_argument(
            "-le",
            "--length_episodes",
            help="length of episodes to learn from",
            default=100,
            type=int,
        )
        self.parser.add_argument(
            "-ce",
            "--checkpoint_episodes",
            help="number of episodes after which a checkpoint occurs",
            default=100,
            type=int,
        )
        self.parser.add_argument(
            "-ex",
            "--extract_all_states",
            help="extract all visited states",
            default=False,
            action="store_true",
        )
        self.parser.add_argument(
            "-es", "--eps_start", help="initial value of epsilon", default=1, type=float
        )
        self.parser.add_argument(
            "-ee", "--eps_end", help="final value of epsilon", default=0.001, type=float
        )
        self.parser.add_argument(
            "-ed",
            "--eps_decay",
            help="decay value for epsilon",
            default=0.999,
            type=float,
        )
        self.parser.add_argument(
            "-bfs",
            "--buffer_size",
            help="size of the replay buffer",
            default=10000,
            type=int,
        )
        self.parser.add_argument(
            "-nsv",
            "--number_state_variables",
            help="number of state variables used to describe a single state",
            default=1,
            type=int,
        )
        self.parser.add_argument(
            "-bs", "--batch_size", help="size of learning batch", default=64, type=int
        )
        self.parser.add_argument(
            "-bns",
            "--best_network_score",
            help="best possible score achieved in pre training so far",
            default=-float("inf"),
            type=float,
        )
        self.parser.add_argument(
            "-pnn",
            "--pretrained_neural_network",
            help="path to pre-trained neural network",
            default=None,
            type=str,
        )
        self.parser.add_argument(
            "-g", "--gamma", help="gamma", default=0.99, type=float
        )
        self.parser.add_argument(
            "-t", "--tau", help="softupdate factor tau", default=0.001, type=float
        )
        self.parser.add_argument(
            "-lr", "--learning_rate", help="learning rate", default=5e-4, type=float
        )
        self.parser.add_argument(
            "-ue", "--update_every", help="update every", default=10, type=float
        )
        self.parser.add_argument(
            "-nnf",
            "--neural_network_file",
            help="path to network file to use",
            default="networks.fcn",
            type=str,
        )
        self.parser.add_argument(
            "-nnw",
            "--neural_network_weights",
            help="weights for network init",
            nargs="+",
            default=None,
            type=int,
        )
        self.parser.add_argument(
            "-gid", "--gpu_id", help="id of gpu to use", default=0, type=int
        )
        self.parser.add_argument(
            "-pef",
            "--policy_extraction_frequency",
            help="fixed number of steps after which policy is extracted",
            default=0,
            type=int,
        )
        self.parser.add_argument(
            "-pr",
            "--positive_reward",
            help="reward given when successfull",
            default=100,
            type=int,
        )
        self.parser.add_argument(
            "-nr",
            "--negative_reward",
            help="reward given when failing",
            default=-100,
            type=int,
        )
        self.parser.add_argument(
            "-sr",
            "--step_reward",
            help="reward given when neither winning nor losing",
            default=0,
            type=int,
        )

        self.dqn_keys = [
            "hermes_name",
            "seed",
            "extract_all_states",
            "num_episodes",
            "length_episodes",
            "checkpoint_episodes",
            "eps_start",
            "eps_end",
            "eps_decay",
            "buffer_size",
            "batch_size",
            "best_network_score",
            "pretrained_neural network",
            "gamma",
            "tau",
            "learning_rate",
            "update_every",
            "neural_network_file",
            "neural_network_weights",
            "gpu",
            "gpu_id",
            "policy_extraction_frequency",
            "extract_replay_buffer",
        ]
        self.spec_keys = [self.dqn_keys]

    def parse(self, input=None):
        """
        Parse the given input and return the arguments.
        Args:
            input (string, optional): input to parse istead of stdargs. Defaults to None.

        Returns:
            Namespace: arguments.
        """
        if input == None:
            self.args = self.parser.parse_args()
        else:
            self.args = self.parser.parse_args(input)
        return self.args

    def get_namespaces(self):
        """
        Returns the namespaces separetely as defined by the keys in the init method.

        Returns:
            Namespace: arguments.
        """
        dicts = [{} for _ in range(len(self.spec_keys))]

        args_dict = vars(self.args)
        for key in args_dict:
            for i, keys in enumerate(self.spec_keys):
                if key in keys:
                    dicts[i][key] = args_dict[key]

        namespaces = []
        for i in range(len(self.spec_keys)):
            ap = argparse.ArgumentParser()
            d = dicts[i]
            ap.set_defaults(**d)
            namespaces.append(ap.parse_args([]))

        return namespaces



#! /usr/bin/env python

import argparse
import sys
import os
from pathlib import Path
import pathlib

import logging

logging.basicConfig(format="[%(asctime)s] %(message)s", level=logging.INFO)

path_to_hermes = pathlib.Path(__file__).parent.resolve()
sys.path.append(str(path_to_hermes))
from .core.execution_file import ExecutionFile
from .core.execution import Execution
from .core import settings as settings
from .core import bot as bot


def bot_handler(args):
    if args.register:
        handler = bot.Bot_handler()
        handler.register()


def execute_file(args):
    """Executes the instructions in a .hermes file given via args.execution_file

    :param args: Namespace returned by hermes parser
    """
    logging.info("Executing from hermes execution file %s", str(args.execution_file))
    execution_file = ExecutionFile(
        args.execution_file, args.comment if args.comment else [], args.root
    )
    execution = Execution(execution_file, args.name, args.debug)
    if args.compile_only:
        execution.print()
    else:
        execution.run()


def single_execution(args):
    """Executes a single command given to hermes via args.executable

    :param args: Namespace returned by hermes parser
    """
    # TODO input validation
    logging.info('Perfoming single execution of "%s"', args.executable)
    raise NotImplementedError


def execute(args):
    """Function called by exec command checks if a .hermes file or a command is given.

    :param args: Namespace returned by hermes parser
    """
    if args.executable:
        single_execution(args)
    elif args.execution_file:
        execute_file(args)


def create_hermes_execution_file(args):
    """Create a barebone execution file in args.target_directory with name args.filename.
        if script names are given in args.scripts, respective template exec lines are created.

    :param args: Namespace returned by hermes parser
    """
    # TODO input validation
    logging.info('Creating barebone hermes execution file "%s"', args.target_directory)
    args.scripts = args.scripts if args.scripts else ["your_script.py"]
    args.scripts = [s + ".py" if not ".py" in s else s for s in args.scripts]
    n_scripts = len(args.scripts)
    path = os.path.join(args.target_directory, args.filename)
    with open(path, "w") as f:
        ExecutionFile.write_pre(f)
        ExecutionFile.write_std(f)
        ExecutionFile.write_exec(args.scripts, [{}] * n_scripts, [{}] * n_scripts, f=f)


def detailed_help(args):
    """Prints a deatiled help. Used by the help command. If a
    command is specified print help only for that command. Else
    print help for all commands.

    :param args: Namespace returned by hermes parser
    """
    if args.command == "ALL":
        for command in args.commands:
            print('Help for "hermes %s":' % command)
            args.commands[command].print_help()
            print("")
    elif args.command in args.commands:
        args.commands[args.command].print_help()
    else:
        print("Unknown Command: %s" % args.command)

def _git_is_available() -> bool:
    try:
        settings.get_repo_path()
        return True
    except RuntimeError:
        logging.warning("Git is not available.")
        return False


def main():
    # system information
    git_is_available = _git_is_available()

    """Entry point for hermes. This function is called when execuinh "hermes [...]" in console."""
    # root parser
    hermes_parser = argparse.ArgumentParser(
        prog="hermes",
        description="A command line tool for organized execution of "
        "experiments in form of python scripts",
    )
    hermes_parser.add_argument(
        "-ho",
        action="store_true",
        help="Print help for hermes options used by .hermes files and exit",
    )
    subparsers = hermes_parser.add_subparsers(
        help="hermes commands, exactly one is required", dest="subparser_name"
    )

    ## default hermes settings
    hermes_parser.set_defaults(**settings.DEFAULT_SETTINGS)

    # create subparser for single execution and file execution
    ## hermes exec
    exec_parser = subparsers.add_parser(
        "exec",
        help="Used to manage execution of either a given call to a python script (single execution), or a hermes execution file",
    )
    ## make sure EXACTLY ONE of the execution options is used
    exec_types = exec_parser.add_mutually_exclusive_group(required=True)
    exec_types.add_argument(
        "-e",
        "--executable",
        help="Provided a script call in quotes, perform single execution",
    )
    exec_types.add_argument(
        "-f",
        "--execution_file",
        type=Path,
        help="Provided a path to a hermes execution file, follow the instructions it gives",
    )
    exec_parser.add_argument(
        "-n",
        "--name",
        type=str,
        required=True,
        help="Give the experiment(s) you are about to execute a name to be included in their identifier",
    )
    exec_parser.add_argument(
        "-c",
        "--comment",
        type=str,
        action="append",
        help="Optional comment that is going to be stored in each hermes.info file created as a result of this execution",
    )
    exec_parser.add_argument(
        "--debug",
        action="store_true",
        help="Enables Debug mod. Storage is skipped in order to prevent cluttered experiment directories",
    )
    exec_parser.add_argument(
        "--compile_only",
        action="store_true",
        help="Only compile the hermes file and show what would normally be executed without actually executing anything",
    )
    exec_parser.add_argument(
        "--root",
        type=Path,
        help="""
            Per default, Hermes uses the git repository root as the root path for IO operations, e.g. for loading hermes settings.
            This argument allows you to override the root path. If git is not available, it is strictly required, otherwise it is optional.
        """,
        required=not git_is_available,
        default=None
    )

    ## set function to be called for execution command
    exec_parser.set_defaults(func=execute)

    # create subparser for barebone execution file creation
    ## hermes create
    create_parser = subparsers.add_parser(
        "create", help="Used for creation of barebone execution files"
    )
    create_parser.add_argument("filename", type=str)
    create_parser.add_argument(
        "-s",
        "--scripts",
        required=False,
        action="append",
        help="Optional list of script names to include in the barebone file",
    )
    create_parser.add_argument(
        "-t",
        "--target-directory",
        required=False,
        action="store",
        default="./",
        help="Optional path to target directory, by default the current working directory",
    )

    ## set function to be called for create command
    create_parser.set_defaults(func=create_hermes_execution_file)

    # bot subparser for telegram bot functions
    ## hermes bot
    bot_parser = subparsers.add_parser("bot", help="Interface to Hermes Telegram Bot")
    bot_parser.add_argument(
        "--register",
        action="store_true",
        help="Register for Hermes Telegram Bot. Your bot id will be read from the hermes settings file.",
    )
    bot_parser.set_defaults(func=bot_handler)

    # create subparser for detailed help
    help_parser = subparsers.add_parser(
        "help", help="Display a detailed help for all commands or a given command"
    )
    help_parser.add_argument("command", type=str, default="ALL", nargs="?")
    help_parser.set_defaults(func=detailed_help)
    help_parser.set_defaults(commands={"exec": exec_parser, "create": create_parser})

    # parse args
    args = hermes_parser.parse_args()

    if args.ho:
        print("Hermes Options usage")
        ExecutionFile.option_parser.print_help()
        exit()

    # no command given
    if not hasattr(args, "func"):
        hermes_parser.print_help()
        exit()

    # run selected command
    logging.info(
        "Launching Hermes with experiments path set to %s", args.path_to_experiments
    )
    args.func(args)


if __name__ == "__main__":
    main()



import re
import itertools
from collections import deque

def _sprit(s, delim):
    return [x.lstrip().rstrip() for x in s.split(delim)]


def _parse_list(lst_str):
    lst_str = lst_str.lstrip().rstrip()
    assert lst_str[0] == "[" and lst_str[-1] == "]"
    contents = _sprit(lst_str[1:-1], ",")
    return contents

_kwarg_pattern = re.compile("([^(=|,|\(|\))]+)=(\[[^(=|\(|\))]+\])")
_exec_line_pattern = re.compile("^\[\[.*\],\[.*\]\]$")
_directive_pattern = re.compile("^[a-zA-Z]+\(.*\)")


def outer_split(contents, sep=","):
    """
    Used to parse expressions of the form
    (A, B, C, ...)
    where A, B, C as expressions can themselves contain brackets and commas within these brackets.

    The point is that we do not want to split at commas within brackets, just at commas that are on the
    outermost layer.
    """

    OPENING_BRACKETS = {
            "(" : ")",
            "{" : "}",
            "[" : "]"
    }

    CLOSING_BRACKETS = {value : key for key, value in OPENING_BRACKETS.items()}


    brackets = deque() # for pushing open brackets onto a stack
    buf = [] # for buffering all characters of one outer split element
    split_result = [] 
    for char in contents:
        if len(brackets) == 0:
            if char == sep:
                split_result.append("".join(buf).lstrip().rstrip())
                buf = []
            else:
                assert char not in CLOSING_BRACKETS, "Found %s but there is no %s left to close." % (char, CLOSING_BRACKETS[char])
                buf.append(char)
                if char in OPENING_BRACKETS:
                    brackets.append(char)
        else:
            buf.append(char)
            if char in OPENING_BRACKETS:
                brackets.append(char)
            elif char in CLOSING_BRACKETS:
                opener = brackets.pop()
                expected_closer = OPENING_BRACKETS[opener]
                assert char == expected_closer, "Expected %s to be closed with %s, but found %s instead." % (opener, expected_closer, char)

    assert len(brackets) == 0, "Unexpected end of expression, not all brackets (%s) closed in:\n%s." % (",".join(b for b in brackets), contents)

    split_result.append("".join(buf).lstrip().rstrip())
    return split_result


class Expression:

    def __init__(self):
        self.needs_compilation = True

    def resolve(self):
        raise NotImplementedError()

    def _match(self, line):
        raise NotImplementedError()

    def match(self, line):
        if self._match(line):
            self.line = line
            return True
        else:
            return False

class ExecLine(Expression):

    def __init__(self):
        self.needs_compilation = False

    def _match(self, line):
        match = _exec_line_pattern.match(line)
        if match:
            self.line = line
            return True
        else:
            return False

    def resolve(self, line):
        return [self.line], False

class Directive(Expression):

    def __init__(self, identifier):
        self.identifier = identifier
        self.pattern = re.compile("^%s\((.*)\)$" % self.identifier)

    def _match(self, line):
        match = self.pattern.match(line)
        if match == None:
            return False
        else:
            self.contents = match.group(1)
            self.contents = outer_split(self.contents)
            return True

    def resolve(self):
        raise NotImplementedError()

    def initial_parse(self, exec_line):

        kwargs = [_kwarg_pattern.match(itm) for itm in self.contents]

        args = [self.contents[i] for i, kwarg in enumerate(kwargs) if kwarg is None]
        assert len(args) >= 1

        kwargs = [(itm.group(1), itm.group(2)) for itm in kwargs if itm is not None]
        kwargs = {key.lstrip().rstrip() : _parse_list(val) for key, val in kwargs}
        return args, kwargs

class Zip(Directive):

    def __init__(self):
        super().__init__("ZIP")

    def resolve(self, exec_line):
        args, kwargs = self.initial_parse(exec_line)

        assert len(args) > 0

        lengths = [len(lst) for lst in kwargs.values()]
        assert len(set(lengths)) == 1, "All ZIP-argument lists must be of same length"
        lst_len = lengths[0]
        
        lines = []
        filling = lambda idx, formatters : {fs : kwargs[fs][idx] if fs in kwargs else "{%s}" % fs for fs in formatters}
        for arg in args:
            formatters = set(re.findall("\{([a-zA-Z]+)\}", arg))
            lines.extend([arg.format(**filling(idx, formatters)) for idx in range(lst_len)])

        del self.contents
        return lines, True


class Prod(Directive):

    def __init__(self):
        super().__init__("PROD")

    def resolve(self, exec_line):
        args, kwargs = self.initial_parse(exec_line)

        assert len(args) > 0

        kwargs = [[(name, val) for val in kwargs[name]] for name in kwargs]

        lines = []
        for arg in args:
            lines.extend([arg.format(**{name : val for name, val in comb}) for comb in itertools.product(*kwargs)])

        del self.contents
        return lines, True

class _Directives(list):

    def __init__(self):
        super().__init__([ExecLine(), Prod(), Zip()])

    def resolve(self, exec_line):
        for i, expr in enumerate(self):
            if expr.match(exec_line):
                assert not any([other_expr.match(exec_line) for other_expr in self[(i+1):]])
                return expr.resolve(exec_line)
        return [exec_line], False

ExecDirectives = _Directives()

if __name__ == "__main__":
    test = "ZIP([[./test.py, -alpha {a}, -beta {b}],[-ca test.py]], [[./foo.py, -alpha {a}, -beta {b}],[-ca foo.py]], a=[0.05, 0.01, 0.1], b=[100, 500, 1000])"
    ZIP = Zip()
    assert ZIP.match(test)
    lines, _ = ZIP.resolve(test)
    [print("ZIP", line) for line in lines]

    test = "PROD([[./test.py, -alpha {a}, -beta {b}],[-ca test.py]], a=[0.05, 0.01, 0.1], b=[100, 500])"
    PROD = Prod()
    assert PROD.match(test)
    lines, _ = PROD.resolve(test)
    [print("PROD", line) for line in lines]


    test = "ZIP([[./test.py, -a {a}, -b {b}],[-ca test.py]], [[./foo.py, -a {a}, -b {b}],[-ca foo.py]], a=[0.05, 0.01], b=[100, 500])"
    test = "ZIP(PROD([[./test.py, -a {a}, -b {b} --seed {seed}],[-ca test.py]], [[./foo.py, -a {a}, -b {b}, --seed {seed}],[-ca foo.py]], a=[0.05, 0.01], b=[100, 500]), seed=[0,1,2])"
    assert ZIP.match(test)

    for x in ZIP.contents:
        print(x)

    lines, _ = ZIP.resolve(test)
    [print("ZIP 2", line) for line in lines]

    for i, line in enumerate(lines):
        assert PROD.match(line)
        print("PROD 2.%d" % i)
        prod_lines,_ = PROD.resolve(line)
        [print(pl) for pl in prod_lines]



import torch
import numpy as np
import numpy
import logging

logging.basicConfig(format="[%(asctime)s] %(message)s", level=logging.INFO)


class Pseudo_agent:
    """
    Pseudo agent for evaluation - assumes the agent is based on q-values.
    Can either work with discrete state variables or with the environemnt.
    """

    def __init__(self, network, name=None):
        self.network = network
        if name is None:
            self.name = "Pseudo Agent"
        else:
            self.name = name

    def get_name(self):
        """
        get the chosen name

        Returns:
            string: name
        """
        return self.name

    def act_with_env(self, env):
        """same as act, but with env option only

        Args:
            env ([type]): [description]

        Returns:
            [type]: [description]
        """
        return self.act(env=env)

    def act(self, state=None, env=None, applicable_actions=None):
        """
        method to decide for an action.
        Either specify state (and possible the available actions) or environment.

        Args:
            state (array, optional): values representing the state. Defaults to =None.
            env (Environment, optional): environment to extract the state from. Defaults to =None.
            blocked_actions (array, optional): ids if the actions that aren't applicable in the current state. Defaults to = None. (which assumes all actions are applicable)

        Returns:
            int: id of action to be applied
        """
        x = self.q_vals(state, env, applicable_actions)

        return np.argmax(x)

    def q_vals(self, state=None, env=None, applicable_actions=None):
        """return the approximation of the q-values

        Args:
            state (array, optional): values representing the state. Defaults to =None.
            env (Environment, optional): environment to extract the state from. Defaults to =None.
            blocked_actions (array, optional): ids if the actions that aren't applicable in the current state. Defaults to = None. (which assumes all actions are applicable)

        Returns:
            array: q-values
        """
        assert not (
            env is None and state is None
        ), "You must specify at least one way for the agent to interact with the environment"
        assert not (
            env is not None and state is not None
        ), "You must either specifiy a state or an environment"
        assert not (
            env is not None and applicable_actions is not None
        ), "You must either specify the environment or discrete representations, don't mix both ways."

        if state is None:
            # if you want to use another method to call the state, change the next line
            state = env.get_state()
            applicable_actions = env.applicable_actions()

        # check and cast state type
        if type(state) == np.ndarray:
            state = torch.from_numpy(state)
        else:
            if type(state) == torch.Tensor:
                pass
            else:
                state = torch.tensor(state)

        with torch.no_grad():
            x = self.network.forward(state)
            x = x.numpy()

            if applicable_actions is not None:
                for i in range(len(x)):
                    if i not in applicable_actions:
                        x[i] = -float("inf")

            return x

    def report(self, one=None, two=None, three=None, four=None, five=None):
        pass

    def finish(self):
        pass



import os
from datetime import datetime
import sys

import torch
import numpy as np
import matplotlib.pyplot as plt
import logging

from pathlib import Path
import pathlib

# path_to_hermes = pathlib.Path(__file__).parent.resolve().joinpath('..')
# sys.path.append(str(path_to_hermes))
from . import util


class Result(dict):
    """
    A class for storing results, i.e. data that is generated during the execution of some script/experiment.

    This class inherits from :python:`dict` and provides useful methods for
    - storing data in the underlying dictionary
    - saving the dictionary to disk
    - accessing stored data
    - plotting stored data

    Note that these plots are meant for rapid prototyping/debugging and are neither very pretty, nor camera-ready as of yet.
    """

    def __init__(self, name, path=None, dictionary=None):
        if dictionary is not None:
            super().__init__(dictionary)
        else:
            super().__init__()
        self.name = name
        self.path = path

    @property
    def dictionary(self):
        return {key: val for key, val in self.items()}

    @classmethod
    def new(cls, name):
        """
        Create a new result and give it a name

        :param name: the result name
        :type name: str

        :return: the result
        :rtype: Result
        """
        res = cls(name)
        return res

    @classmethod
    def from_namespace(cls, name, namespace):
        """
        Create a new result from an existing namespace

        :param name: the result name
        :type name: str

        :param namespace: the namespace to use
        :type namespace: namespace
        """
        d = vars(namespace)
        cls(name, dictionary=d)

    def save(self, directory=None):
        """
        Save the result to disk. Uses :python:`torch.save` internally.

        :param directory: the directory, where to save the result. Set to :python:`None`, if the result already has a path and you wish to overwrite,
        default to :python:`None`. Note that you will have to specify a path on every first save.
        :type directory: Union[str, os.PathLike]
        """
        if directory:
            save_at = os.path.join(directory, self.name + ".result")
        else:
            save_at = self.path
        if save_at is None:
            raise ValueError(
                "The result has not been assigend a path yet, so you have to specify a directory where to save it"
            )
        torch.save(self.dictionary, save_at)

    @classmethod
    def load(cls, path):
        """
        Load a result from disk

        :param path: the path to a result file
        :type str:

        Some things you should know about this method:
        - stored scalar lists are converted to numpy arrays on load
        - internally it uses :python:`torch.load`

        :return: the result
        """
        path = Path(path)
        assert path.exists() and path.is_file() and path.suffix == ".result"
        dictionary = torch.load(path)
        name = path.stem
        res = cls(name, dictionary=dictionary, path=path)
        for key in res:
            if isinstance(res[key], list) and key != "tags":
                try:
                    res[key] = np.array(res[key])
                except:
                    logging.warning("Unable to convert list %s to numpy" % key)
        return res

    def store_scalar(self, value, key):
        """
        Store a scalar, or list/iterable of scalars associated with some key, i.e. name / string identifier.

        :param value: The value to store
        :type value: Union[bool, int, float, Iterable[Union[bool, int, float]]]

        :param key: The key under which the scalar is stored
        :type key: str

        If the key is new, a list entry in the result dictionary is associated with it,
        otherwise the scalar(s) is (are) appended to the key's list.
        """
        if not hasattr(value, "__iter__"):
            value = [value]
        if key in self.keys():
            self[key].extend(value)
        else:
            self[key] = list(value)

    def store_array(self, value, key):
        """
        Similar to :func:`~Result.store_scalar`, but store lists of values, without flatting them into
        one dimension. After loading this list, it will be a 2d numpy :type:numpy.ndarray.

        :param value: The array to store
        :type value: Union[List[Any], numpy.ndarray]

        :param key: The key under which the array is stored
        :type key: str
        """
        value = np.array(value)
        if key in self.keys():
            self[key].append(value)
        else:
            self[key] = [value]

    def plot_scalar(self, key, smoothing=50, show=True, ax=None):
        """
        Plot the array of scalars stored under a specific key

        :param key: the scalars key
        :type key: str

        :param smoothing: Size of a smoothing window applied to the scalars before plotting, defaults to 50
        :type smoothing: int, optional

        :param show: Whether to immediately show the plot, i.e. call :python:`plt.show`, defaults to True
        :type show: bool, optional

        :param ax: An axis to to plot on, defaults to None
        :type ax: matplotlib.axis, optional
        """
        x = util._moving_average(self[key], smoothing)
        if not ax:
            plt.plot(x, label=key)
            if show:
                plt.legend()
                plt.show()
        else:
            ax.plot(x)



import numpy as np
import torch
from collections import deque
import random



class ReplayBuffer:
    # initialize the Replay buffer
    # Fix the seed for random sampling through the Replay Buffer
    # Batch size defines number of samples drawn at each learning operation
    # buffer is the actual buffer
    def __init__(self, buffer_size, num_state_variables,  state_variable_sizes, batch_size, seed = 0, device = None):
        assert (num_state_variables) == len(state_variable_sizes), "The provided number of state variables and the provided dimensions differ"
        self.batch_size = batch_size
        self.buffer_size = buffer_size
        self.num_state_variables = num_state_variables
        self.state_variable_sizes = state_variable_sizes
        # determine the size of the state buffers

        # initialize a full buffer - store all parts apart from each other for faster sampling
        self.states = [0 for _ in range(num_state_variables)]
        self.next_states = [0 for _ in range(num_state_variables)]
        for i,variable_size in enumerate(state_variable_sizes):
            dimension = tuple([buffer_size] + list(variable_size))
            self.states[i] = np.zeros(dimension)
            self.next_states[i] = np.zeros(dimension)

        self.actions = np.zeros(buffer_size)
        self.rewards = np.zeros(buffer_size)
        self.dones = np.zeros(buffer_size)
        # buffer counter
        self.counter = 0
        self.filled = False
        # use only numpys random
        self.seed = np.random.seed(seed)

        # if no external device is give, use CPU
        if device == None:
            self.device = torch.device("cpu")
        else:
            self.device = device

    # add samples to the buffer
    # transfer the done value to an integer
    # we assume, that a state is given through several parts, so you always need to provide a list
    def add(self, state, action, reward, next_state, done):
        assert len(state) == self.num_state_variables, "Provided state did not contain all state variables"
        assert len(next_state) == self.num_state_variables, "Provided next_state did not contain all state variables"

        # transformation to integer
        if done:
            done_value = 1
        else:
            done_value = 0

        # fill buffer
        for i, (state_variable, next_state_variable) in enumerate(zip(state,next_state)):
            self.states[i][self.counter] = state_variable
            self.next_states[i][self.counter] = next_state_variable

        self.actions[self.counter] = action
        self.rewards[self.counter] = reward
        self.dones[self.counter] = done_value

        # increase counter (and reset, if nessary)
        self.counter += 1
        if self.counter == self.buffer_size:
            self.counter = self.counter % self.buffer_size
            self.filled = True

    # sample from the database
    # the samples later need to be split into tensors of each part of the samples
    # thus, collects a sample and writes every part of the sample in the corresponding list
    # afterwards transforms this lists into tensors and returns them
    def sample(self):
        sample_ids = np.random.randint(0, self.__len__(), self.batch_size)
        states = [0 for _ in range(self.num_state_variables)]
        next_states = [0 for _ in range(self.num_state_variables)]
        for i in range(self.num_state_variables):
            tmp = self.states[i][tuple([sample_ids])]
            states[i] = torch.from_numpy(tmp).float().to(self.device)
            tmp =  self.next_states[i][tuple([sample_ids])]
            next_states[i] = torch.from_numpy(tmp).float().to(self.device)

        actions = self.actions[tuple([sample_ids])]
        rewards = self.rewards[tuple([sample_ids])]
        dones = self.dones[tuple([sample_ids])]

        # next state and state have already been transformed
        actions = torch.from_numpy(actions).long().to(self.device)
        rewards = torch.from_numpy(rewards).float().to(self.device)
        dones = torch.from_numpy(dones).float().to(self.device)
        return [states, actions, rewards, next_states, dones]

    # depending on whether the buffer is completely filled or whether is still about to be filled
    def __len__(self):
        if self.filled:
            return self.buffer_size
        else:
            return self.counter


import numpy as np
import torch
from collections import OrderedDict
from dataclasses import dataclass
from typing import Any, Tuple, Union, List


@dataclass
class Variable:
    """
    Represents objects that can be stored, e.g. state or reward
    """

    name: str
    dtype: torch.dtype = torch.float32 # data type, should be a torch data type
    shape: Tuple[Tuple[int]] = ((1,),) # Nesting is possible, e.g. if state consists of two tensors with different shapes

class Batch(OrderedDict):

    def __init__(self, **kwargs):
        super().__init__(kwargs)

    def unpack(self):
        return self.values()

class Buffer:
    """
    Implements a Buffer for exactly one type of variable
    The variable may however consists of multiple components with different shapes.
    """
    def __init__(self, size, variable):
        self.variable = variable
        self.buffers = [np.zeros((size,) + tuple(shape)) for shape in self.variable.shape]

    def set(self, idx, variable_instance):
        if len(self.variable.shape) > 1: # more than one component
            for i, component_instance in enumerate(variable_instance):
                self.buffers[i][idx] = component_instance
        else: # only one component
            self.buffers[0][idx] = variable_instance

    def get(self, idx, device):
        if len(self.variable.shape) > 1: # more than one component
            return [torch.from_numpy(self.buffers[i][idx]).to(self.variable.dtype).to(device)
                    for i in range(len(self.variable.shape))]
        else: # only one component
            return torch.from_numpy(self.buffers[0][idx]).to(self.variable.dtype).to(device)


class ReplayBuffer:
    """
    A more flexible replay buffer implementation for use with pytorch

    Can be useful when implementing e.g. n-step returns, when storing continuous actions that have some
    fixed shape, when there is more than one reward signal, [...].
    """


    def __init__(self,
                 size: int,
                 batch_size: int,
                 variables: List[Variable],
                 seed: Any = 0,
                 device=None):
        """

        :param size: Buffer size, i.e. maxmimal number of entries
        :type size: int

        :param batch_size: size of sampled batches
        :type batch_size: int

        :param variables:
        :type variables: List[Variable]

        :param seed:
        :type seed: Any

        :param device:
        """
        self.device = device
        self.size = size
        self.batch_size = batch_size
        self.variables = variables
        self.seed = seed
        self.rng = np.random.default_rng(seed)

        self.buffers = {variable.name : Buffer(size, variable) for variable in variables}
        self.counter = 0
        self.num_stored = 0

    def add(self, **variable_instances):
        for variable_name, variable_instance in variable_instances.items():
            self.buffers[variable_name].set(self.counter, variable_instance)

        self.counter = (self.counter + 1) % self.size
        self.num_stored = min(self.size, self.num_stored + 1)

    def __len__(self):
        return self.num_stored

    def sample(self):
        sample_ids = self.rng.integers(0, len(self), self.batch_size)
        return Batch(**{var.name : self.buffers[var.name].get(sample_ids, self.device) for var in self.variables}).unpack()


# usage example
class StandardReplayBuffer(ReplayBuffer):

    def __init__(self, state_variable_sizes, *args, **kwargs):
        variables = []
        variables.append(Variable("state", shape=tuple(tuple(svs) for svs in state_variable_sizes)))
        variables.append(Variable("action", dtype=torch.long))
        variables.append(Variable("reward"))
        variables.append(Variable("next_state", shape=variables[0].shape))
        variables.append(Variable("done"))
        super().__init__(*args, variables, **kwargs)


# usage example continued
if __name__ == "__main__":
    buf = StandardReplayBuffer([[5,12],[1]], 100, 4)
    for i in range(1000):
        buf.add(state=[np.random.rand(5, 12), 3], action=4, reward=0, next_state=[np.random.rand(5,12), 2], done=0)

    batch = buf.sample()
    state, action, reward, next_state, done = batch.unpack()
    print(state)



from __future__ import annotations

import torch
import numpy as np
import leb128
import json
import typing as t
import os
import copy


class UnsupportedLayerError(Exception):
    pass


def save_network(model, path):
    torch.save(model.state_dict(), path)


def load_network(model, path):
    d = torch.load(path, map_location=model.device)
    model.load_state_dict(d)
    model.eval()


def _make_if_necessary(p):
    if not os.path.exists(p):
        os.mkdir(p)


def _moving_average(x, w):
    return np.convolve(x, np.ones(w), "valid") / w


def _is_numeric(val):
    return isinstance(val, int) or isinstance(val, float)


# copyright TorchSharp
# https://github.com/dotnet/TorchSharp
def _elem_type(t):
    dt = t.dtype

    if dt == torch.uint8:
        return 0
    elif dt == torch.int8:
        return 1
    elif dt == torch.int16:
        return 2
    elif dt == torch.int32:
        return 3
    elif dt == torch.int64:
        return 4
    elif dt == torch.float16:
        return 5
    elif dt == torch.float32:
        return 6
    elif dt == torch.float64:
        return 7
    elif dt == torch.bool:
        return 0
    elif dt == torch.bfloat16:
        return 15
    else:
        return 4711


# copyright TorchSharp
# https://github.com/dotnet/TorchSharp
def _write_tensor(t, stream):
    stream.write(leb128.u.encode(_elem_type(t)))
    stream.write(leb128.u.encode(len(t.shape)))
    for s in t.shape:
        stream.write(leb128.u.encode(s))
    stream.write(t.numpy().tobytes())


# copyright TorchSharp
# https://github.com/dotnet/TorchSharp
def _save_state_sharp_dict(sd, stream):
    """
    Saves a PyToch state dictionary using the format that TorchSharp can
    read.
    :param sd: A dictionary produced by 'model.state_dict()'
    :param stream: An write stream opened for binary I/O.
    """
    stream.write(leb128.u.encode(len(sd)))
    for entry in sd:
        stream.write(leb128.u.encode(len(entry)))
        stream.write(bytes(entry, "utf-8"))
        _write_tensor(sd[entry], stream)


def save_network_as_dat(model, path):
    assert path.endswith(".dat"), "This method can only be used to save a .dat file"
    f = open(path, "wb")
    tmp_model = copy.deepcopy(model)
    _save_state_sharp_dict(tmp_model.to("cpu").state_dict(), f)
    f.close()


# Copyright (C) 2021, Saarland University
# Copyright (C) 2021, Maximilian Köhl <koehl@cs.uni-saarland.de>
def _dump_layer(name: str, layer: torch.nn.Module) -> t.Any:
    result = {"name": name, "kind": layer.__class__.__name__}
    if isinstance(layer, torch.nn.Linear):
        result["inputSize"] = layer.in_features
        result["outputSize"] = layer.out_features
        result["hasBiases"] = getattr(layer, "bias", None) is not None
        result["weights"] = layer.weight.tolist()
        result["biases"] = layer.bias.tolist()
    elif isinstance(layer, torch.nn.ReLU):
        pass
    elif isinstance(layer, torch.nn.CELU):
        result["alpha"] = layer.alpha
    else:
        raise UnsupportedLayerError(f"layer of type {type(layer)} not supported")
    return result


# Copyright (C) 2021, Saarland University
# Copyright (C) 2021, Maximilian Köhl <koehl@cs.uni-saarland.de>
def _dump_layers(net: torch.nn.Sequential) -> t.Any:
    return [_dump_layer(name, layer) for name, layer in net.named_children()]


# Copyright (C) 2021, Saarland University
# Copyright (C) 2021, Maximilian Köhl <koehl@cs.uni-saarland.de>
def dump_nn(net: torch.nn.Module) -> str:
    if not isinstance(net, torch.nn.Sequential):
        result = net.forward(_FakeTensor())
        assert isinstance(
            result, _FakeTensor
        ), "the result of forwarding should be a `_FakeTensor`"
        layers = []
        while result.module is not None:
            layers.append(result.module)
            result = result.parent
        net = torch.nn.Sequential(*reversed(layers))
    return json.dumps({"layers": _dump_layers(net)}, indent=2)


# Copyright (C) 2021, Saarland University
# Copyright (C) 2021, Maximilian Köhl <koehl@cs.uni-saarland.de>
class _FakeTensor:
    parent: t.Optional[_FakeTensor]
    module: t.Optional[torch.nn.Module]

    def __init__(
        self,
        parent: t.Optional[_FakeTensor] = None,
        module: t.Optional[torch.nn.Module] = None,
    ):
        self.parent = parent
        self.module = module

    def __torch_function__(self, func, types, args=(), kwargs=None):
        if func is torch.functional.F.linear:
            (_, weight) = args
            bias = kwargs.get("bias", None)
            out_features, in_features = weight.shape
            module = torch.nn.Linear(in_features, out_features, bias=bias is not None)
            module.weight = weight
            if bias is not None:
                module.bias = bias
            return _FakeTensor(self, module)
        elif func is torch.functional.F.relu:
            return _FakeTensor(self, torch.nn.ReLU())
        elif func is torch.functional.F.celu:
            return _FakeTensor(self, torch.nn.CELU(alpha=kwargs.get("alpha", 1.0)))
        else:
            raise UnsupportedLayerError(
                f"unsupported function {func!r} applied to layer"
            )


def save_network_as_json(model, path):
    """save network as json file

    Args:
        model (network): network to be saved
        path (string): path to be used for saving.
    """
    assert path.endswith(".json"), "This method can only be used to save a .json file"
    f = open(path, "w")
    f.write(dump_nn(model))
    f.close()

def save_dict(d, name):
    """save the given dictionary

    Args:
        d (dictionary): dictionary to be saved
        name (string): name to be used for saving
    """
    with open(name + '.parameters', 'w') as f:
        keys = d.keys()
        keys = sorted(keys)
        num = len(keys)
        for i, key in enumerate(keys):
            f.write(str(key) + ':' + str(d[key]))
            if i != num-1:
                f.write('\n')


def save_namespace(namespace, name):
    """save the given namespace

    Args:
        namespace (namespace): namespace to be saved
        name (string): name to be used for saving
    """
    d = vars(namespace)
    save_dict(d, name)

class ScoreWindow:
    def __init__(self, window_size=100):
        self.window_size = window_size
        self.num = 0
        self.id = 0
        self.window = np.zeros(self.window_size)

    def append(self, element):
        if self.num < self.window_size:
            self.window[self.id] = element
            self.num += 1
        else:
            self.window[self.id] = element
        self.id = (self.id + 1) % self.window_size

    def mean(self):
        assert self.num != 0, "There must be at least one element to define the mean"
        return np.mean(self.window[: self.num])

    def __iter__(self):
        return self.window.__iter__()


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rlmate', 'rlmate.core', 'rlmate.networks']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3.1.24,<4.0.0',
 'leb128>=1.0.4,<2.0.0',
 'matplotlib>=3.4.3,<4.0.0',
 'numpy',
 'pandas>=1.3.4,<2.0.0',
 'python-telegram-bot>=13.6,<14.0',
 'torch==1.9.1']

entry_points = \
{'console_scripts': ['hermes = rlmate.hermes:main']}

setup_kwargs = {
    'name': 'rlmate',
    'version': '0.0.11',
    'description': 'tbd.',
    'long_description': "# RLMate\n\nProject under development. \n\nThe RLMate is a package designed to take over parts of Reinforcement Learning that are needed frequently. \nIt comes together with the command line tool 'Hermes'. \n\nThe idea is to have a tool\n- to organize RL tasks and carry out several scripts,\n- to execute common RL tasks, e.g., plot training curves, \n- have commonly shared code as a python package instead of reimplementing it over and over again.  \n\n\n## Installation\nSimply install the current version with\n* `pip install  rlmate`\n\nFurther instructions will be added when the package involves. \n\n\n## Installation\n\ntodo\n## Command Line Tool\n\nAll available commands can also be seen by using the help function `hermes -h`\n\nAvailable commands:\n* ``exec -f file``: execute the specified Hermesfile. \n* ``exec -e command``: execute the specified command. __Not supported yet!__\n* ``exec -n name``: give the execution file a name to make the results directory more human readable.\n* ``create``: create an example Hermes files\n* ``help command``: show a detailed help of the specified command.\n* `-ho`: show help for hermes options used in .hermes files (also see below) \n\n### Settings\nSettings in this context means something that is not set for a particular execution/hermesfile, but for the __git repository__ that is currently used. \nThe settings are saved in the root if the repository in the ``.hermes_settings`` file. \nThe current settings (with the default value if there is no settings file) are:\n* `path_to_experiment ('./')`: specifies the path where to store the results of the experiment in. Takes the repository root as entry point.\n* `threads (1)`: specifies how many of the specified jobs should be executed parallel.\n\nThe syntax is as follow: `setting:value` and only one setting should be specified per line.\n\n### Hermes-options \nFor each job, different Hermes options can be specified. Currently, these options are:\n* `-ca file`: the `file` is copied to the experiment folder after the execution of the job.\n* `-cb file`: the `file` is copied to the experiment folder before the execution of the job.\n* `-ma file`: the `file` is moved to the experiment folder after the execution of the job.\n* `-l/--log (True | False)`: if `True` the stdout and stderr of the job will not be printed to console, but stored in log files. Defaults to `True`.\n* `-m`: if set, a telegram message will be sent to the chat_ids specified in the `.hermes_settings` file. \n\n\n### Hermesfile\n\nThe Hermesfile specifies the experiments that should be ran. \nHermesfiles have the ending `.hermes`.\nEach file is divided into 3 parts:\n\n#### 1. PRE\nIn the PRE part of the Hermesfile, the settings can be adjustet *for the given Hermesfile only*.\nSo if you generally want to process the files by using two threads, and thereforce specified so in the `.hermes_settings` file, you can change this for the current hermesfile only by specifying it in the PRE section. \nThis is done linewise per setting. \nThe syntax is the same as for the settings file.  \n\n#### 2. STD-H\nSometimes, some Hermes options should be applied to all jobs of the Hermes file. \nHermes-options specified in the STD-H section will be applied to all jobs. \nThe syntax is one Hermes-option per line with `-flag file`\n\n#### 3. STD-E\nAlso, there sometimes are script arguments that should be applied to all jobs while just a few are about to be varied. \nSuch arguments specified in the STD-E section will be applied to all jobs. \nThe syntax in one argument per line, e.g. `-ne 100` or `-n`.\n\nThis standard arguments will be added _before_ the arguments specified in the exec lines. \nNote that this order may be important for your script. \n#### 4. EXEC\nIn this section the jobs are specified with the following syntax:\n```\n[[cmd],[Hermes-options]]\n```\nwhereby the Hermes-options are separated by comma.\nIn each line, one job can be specified. \nAll settings that should be saved automatically must be set in the cmd. \n### Storage\n\nAll files will be saved in the directory `experiments/script_name/uniqe_name`, where\n* `script_name` is the name of the script specified in the job\n* `unique_name` is the unique combination of date and time, commit number, and name that may have been specified with the `-n` flag\n\nand the experiments directory will be created on the first call in the directory as specified in the settings. \n\nIf during the execution of the script new files or directories will be created, they are automatically moved to the storage if they include the `unique_name`. \nHow this name is given to the script will be specified in the next section. \n\nAdditionally, in the future there will be an overview file giving detailed information about all the made experiments. \n\n\n\n### Script Compatibility\nFor your script to be compatible with Hermes, there are basically two things you need to consider:\n1. All the parameters that should be saved must be specified by command line. \nIf you want to save additional parameters, make sure to specify it by archiving the files by using the Hermes-options.\n2. This is the only strong requirement: The first argument of the executed script must be the unique name, that is determined by Hermes. \nYou must not do anything with that identifier except for saving additional files as specified in the Storage section.\n\n",
    'author': 'Timo P. Gros',
    'author_email': 'timopgros@cs.uni-saarland.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import time as t
import numpy as np
import matplotlib.pyplot as plt

MINOR_TRESHOLD = 5

class TimeAnalyzer():

    def __init__(self):
        
        self.last_time = t.time()
        self.checkpoints = {}
        self.times = []
        self.count = []
        
        self.time_log = []
        

    def track(self, name, log = False):
        """
        Track the time elapsed between either
        * the creation time of the TimeAnalyzer, or
        * the last time this method was called
        and the point in time when this method is called.

        :param name: Name the measured elapsed time should be associated with
        :type name: str

        :param log: ?, defaults to False
        :type log: bool
        """
        time = t.time()
        if log:
            self.time_log.append((name, time))
        else:
            current_time = time - self.last_time
            self.last_time = time
            if name in self.checkpoints:
                name_id = self.checkpoints[name]
                self.times[name_id] += current_time
                self.count[name_id] += 1
            else:
                new_id = len(self.checkpoints)
                self.checkpoints[name] = new_id
                self.times.append(current_time)
                self.count.append(1)
            
    def get_times(self):
        return list(self.checkpoints), self.times, self.count
    
    def get_log(self):
        return self.time_log
    
    def save_times(self, file_name = 'time', graphical = False, drop_minors = True):
        relatives = np.zeros(len(self.times))
        names = list(self.checkpoints)

        with open(file_name + '.ta', 'w') as f:
            f.write("Name Time Counter Time/Count\n")
            for i, (name, time, c) in enumerate(zip(names, self.times, self.count)):
                r = time/c
                relatives[i] = r
                f.write(name + " " + str(time) + " " + str(c) + " " + str(float(r)) + "\n")

        ordered_relatives = sorted(relatives)
        ordered_relatives_names = [x for _,x in sorted(zip(relatives, names))]
        ordered_count = [x for _,x in sorted(zip(relatives, self.count))]

        if drop_minors:
            removal = []
            for i, (c, r, n) in enumerate(zip(ordered_count, ordered_relatives, ordered_relatives_names)):
                if c < MINOR_TRESHOLD:
                    removal.append([r,n])

            for r,n in removal:
                ordered_relatives.remove(r)
                ordered_relatives_names.remove(n)

        ordered_times = sorted(self.times)
        ordered_names = [x for _,x in sorted(zip(self.times, names))]

        if graphical:
            plt.figure()
            plt.bar(names, relatives)
            plt.xticks(rotation = 90)
            plt.savefig(file_name + "_time_analysis_relative.png",bbox_inches='tight')
            plt.close()

            plt.figure()
            plt.bar(names, self.times)
            plt.xticks(rotation=90)
            plt.savefig(file_name + '_time_analysis.png',bbox_inches='tight')
            plt.close()

            plt.figure()
            plt.bar(ordered_relatives_names, ordered_relatives)
            plt.xticks(rotation=90)
            plt.savefig(file_name + "_time_analysis_ordered_relative.png",bbox_inches='tight')
            plt.close()

            plt.figure()
            plt.bar(ordered_names, ordered_times)
            plt.xticks(rotation=90)
            plt.savefig(file_name + "_time_analysis_ordered.png",bbox_inches='tight')
            plt.close()


    def save_log(self, file_name = 'time'):
        with open(file_name + '.log', 'w') as f:
            for name, time in self.time_log:
                f.write(name + " " + str(time) + "\n")



import logging
import threading

try:
    from telegram.ext import Updater
    from telegram import Bot, User
    from telegram.ext import CommandHandler
    from telegram.error import (
        TelegramError,
        Unauthorized,
        BadRequest,
        TimedOut,
        ChatMigrated,
        NetworkError,
    )
except ModuleNotFoundError:
    logging.warning(
        "Module telegram was not found, attempting to use any telegram messaging features will break."
    )

import git


def kill_updater(updater):
    updater.stop()
    updater.is_idle = False


class Bot_handler:
    def __init__(self):
        git_repo = git.Repo("./", search_parent_directories=True)
        git_root = git_repo.git.rev_parse("--show-toplevel")

        bot_id = None
        chat_ids = []
        with open(git_root + "/.hermes_settings", "r") as f:
            for line in f:

                if line.startswith("bot_id"):
                    assert bot_id == None, "You can only specify ony bot_id"
                    bot_id = line[7:].replace("\n", "")

                if line.startswith("chat_id"):
                    chat_id = line[8:].replace("\n", "")
                    chat_ids.append(chat_id)

        assert bot_id != None, "you need to specify a bot to send telegram messages"
        if chat_ids == []:
            logging.warning("No chat id was specified. No messages will be sent")

        self.bot_id = bot_id
        self.chat_ids = chat_ids

        self.bot = Bot(self.bot_id)

    def message(self, text):
        for chat_id in self.chat_ids:
            self.bot.send_message(chat_id=chat_id, text=text)

    def picture(self, path="tmp.png", caption=""):
        for chat_id in self.chat_ids:
            self.bot.send_photo(chat_id, photo=open(path, "rb"), caption=caption)

    def register(self):
        updater = Updater(token=self.bot_id)

        def start(update, callback):
            message_id = update.message.chat.id
            res = "Your chat_id is " + str(message_id) + ", please add the line\n"
            res += "chat_id=" + str(message_id)
            res += "\nto your .hermes_settings file"
            update.message.reply_text(res)
            t = threading.Thread(name="killer", target=kill_updater, args=(updater,))
            t.start()

        start_handler = CommandHandler("start", start)
        dispatcher = updater.dispatcher
        dispatcher.add_handler(start_handler)
        updater.start_polling()



import git
import os
from copy import deepcopy
from pathlib import Path


def get_repo_path():
    try:
        repo = git.Repo(".", search_parent_directories=True)
        return repo.working_tree_dir
    # TODO what exception does gitpython throw if git is not available??
    except:
        raise RuntimeError("""
            Cannot use git repository root path as hermes root path, as git is not available!
            Please specify a root path manually with the --root option
        """)


# Dictionary of settings with default value:
DEFAULT_SETTINGS = {"path_to_experiments": "./", "threads": 1}


def add_to_dictionary(dictionary, line, delimiter, line_number):
    if line.startswith("bot_id") or line.startswith("chat_id"):
        return True
    try:
        key, value = line.split(delimiter)
    except:
        print("line " + str(line_number) + ": " + line + " does not fit convention")
        return False
    try:
        t = type(DEFAULT_SETTINGS[key])
    except:
        print("line " + str(line_number) + ": " + key + " setting does not exist")
        return False
    try:
        dictionary[key] = t(value)
    except:
        print(
            "line "
            + str(line_number)
            + ": "
            + value
            + " could not be casted to needed type ("
            + t
            + ")"
        )
        return False

    return True

def get_settings_fp(path_to_root=None):
    if path_to_root is None:
        path_to_root = get_repo_path()
    settings_file = os.path.join(path_to_root, ".hermes_settings")
    return path_to_root, settings_file

def load(path_to_root=None):
    """checks whether the repo has a .hermes_settings file
    loads the given settings if there os one is
    uses the default settings else"""
    settings = deepcopy(DEFAULT_SETTINGS)
    path_to_root, settings_file = get_settings_fp(path_to_root)
    if os.path.exists(settings_file):
        with open(settings_file, "r") as f:
            for i, line in enumerate(f):
                line = line.replace(" ", "")
                line = line.replace("\n", "")
                add_to_dictionary(settings, line, ":", i)

    relative_path = settings["path_to_experiments"]
    absolute_path = Path(path_to_root).absolute()
    settings["path_to_experiments"] = str(Path.joinpath(absolute_path, relative_path))
    return settings


def create_settings_file(path_to_root=None):
    """checks whether the repo has a .hermes_settings file
    gives an error if there is one
    else creates an hermes_settings file with the default values"""
    path_to_root, settings_file = get_settings_fp(path_to_root)
    if os.path.exists(settings_file):
        raise FileExistsError("A hermes settings file already exists!")
    else:
        with open(settings_file, "w") as f:
            for i, key in enumerate(DEFAULT_SETTINGS):
                s = str(key) + ":" + str(DEFAULT_SETTINGS[key])
                if i != len(DEFAULT_SETTINGS):
                    s += "\n"
                f.write(s)



import torch
from torch import nn
import torch.nn.functional as F


cuda = torch.device("cuda")

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")


class Network(nn.Module):
    def __init__(self, device, weights=[15, 64, 64, 9]):
        super().__init__()
        self.device = device
        # self.weights = weights
        self.layers = nn.ModuleList(
            [nn.Linear(weights[i], weights[i + 1]) for i in range(len(weights) - 1)]
        ).to(device)

    def forward(self, x):
        for i in range(len(self.layers) - 1):
            x = self.layers[i].forward(x)
            x = F.relu(x)
        x = self.layers[-1].forward(x)
        return x
